#include "theme.h"

typedef struct ThemeCursor {
	Display *dis;
	Tk_Window tkwin;
	Tk_Window ctkwin;
	Tk_Cursor cursor;
	Pixmap pix;
	Window xWin;
	Window cursorXWin;
	Window rootWin;
	Window mainWin;
	unsigned int width;
	unsigned int height;
	int screen;
	int mapped;
	GC cursorGC;
	int depth;
} ThemeCursor;


static void ThemeCursorEventLeave (ClientData clientData, XEvent *eventPtr) {
	ThemeCursor *wPtr = (ThemeCursor *) clientData;
	
	if (eventPtr->type == LeaveNotify) {
		int detail = eventPtr->xcrossing.detail;
		/*
		fprintf (stderr, "MAIN 1 %d 2 %d 3 %d 4 %d 5 %d 6 %d\n", detail == NotifyNormal, detail == NotifyAncestor, 
			detail == NotifyInferior, detail == NotifyNonlinear, detail == NotifyNonlinearVirtual,
			detail == NotifyVirtual
		);
		*/
		
		if (detail == NotifyAncestor || detail == NotifyNonlinear || detail == NotifyNonlinearVirtual) {
			XUnmapWindow (wPtr->dis, wPtr->cursorXWin);
			wPtr->mapped = 0;
		}
	}
}

static void ThemeCursorEventMotion (ClientData clientData, XEvent *eventPtr) {
	ThemeCursor *wPtr = (ThemeCursor *) clientData;
	int x;
	int y;
	Window wjunk;
	int ijunk;
	unsigned int mask;
	

	if (eventPtr->type != MotionNotify) {
		return;
	}

	XQueryPointer (wPtr->dis, 
		wPtr->mainWin, 
		&wjunk, &wjunk,
		&ijunk, &ijunk,
		&x, &y,
		&mask
	);
	
	if (wPtr->mapped == 0) {
		XMapRaised (wPtr->dis, wPtr->cursorXWin);
		wPtr->mapped = 1;
	}

	XMoveWindow (wPtr->dis, wPtr->cursorXWin, x + 1, y + 1);

	XCopyArea (wPtr->dis, 
		wPtr->pix, wPtr->cursorXWin, 
		wPtr->cursorGC,
		0, 0, 
		wPtr->width, wPtr->height,
		0, 0
	);

	
	XFlush (wPtr->dis);
	/*fprintf (stderr, "FIRED win %ld\n", eventPtr->xmotion.window);*/
}

static void ThemeCursorGetMotionFromAllChildren (ThemeCursor *wPtr, Window win) {
	Window wjunk;
	Window *children = NULL;
	unsigned int nchildren = 0;
	Tk_Window sub_tkwin;
	Window tmpWin;
	unsigned int i;

	if (!XQueryTree (wPtr->dis, win, &wjunk, &wjunk, &children, &nchildren)) {
		return;
	}

	for (i = 0; i < nchildren; ++i) {
		tmpWin = children[i];

		ThemeCursorGetMotionFromAllChildren (wPtr, tmpWin);

		sub_tkwin = Tk_IdToWindow (wPtr->dis, tmpWin);

		if (sub_tkwin == NULL) {
			continue;
		}
		/*fprintf (stderr, "Handling %ld\n", tmpWin);*/
		Tk_DefineCursor (sub_tkwin, wPtr->cursor);
		Tk_CreateEventHandler (sub_tkwin, PointerMotionMask, ThemeCursorEventMotion, (ClientData) wPtr);
	}

	if (children != NULL) {
		XFree (children);
	}
}

static int ThemeCursorCmdWindowCreateCursor (ThemeCursor *wPtr) {
	Window wjunk;
	int ijunk;
	unsigned int uijunk;
	XSetWindowAttributes xattr;
	Pixel winPixel;


	XGetGeometry (wPtr->dis, 
		wPtr->pix, &wjunk,
		&ijunk, &ijunk,
		&wPtr->width, &wPtr->height,
		&uijunk,
		&uijunk
	);

	winPixel = BlackPixel (wPtr->dis, wPtr->screen);
	
	wPtr->cursorXWin = XCreateSimpleWindow (wPtr->dis, 
		wPtr->mainWin,
		1, 1,
		wPtr->width, wPtr->height,
		0,
		winPixel,
		winPixel
	);
	
	xattr.override_redirect = 1;
	/*not currently used*/
	xattr.backing_store = 1;
	XChangeWindowAttributes (wPtr->dis, wPtr->cursorXWin, CWOverrideRedirect, &xattr);

	return 0;
}

static char dummyCursorData[] = {
	"#define d_width 1\n"
	"#define d_height 1\n"
	"static unsigned char d_bits[] = {0x01};\n"
};


int ThemeCursorCmd OBJ_CMD_ARGS {
	char *path;
	char *xpmData;
	Tk_Cursor cursor;
	Tk_Window tkwin;
	ThemeCursor *wPtr = NULL;
	Pixmap pixShape;
	int eventBase;
	int errorBase;
	
	
	/*theme:cursor .path xpmData*/
	if (objc != 3) {
		Tcl_WrongNumArgs(interp, 1, objv, ".path xpmData");
		return TCL_ERROR;
	}
	
	path = Tcl_GetString (objv[1]);
	xpmData = Tcl_GetString (objv[2]);
	tkwin = Tk_NameToWindow (interp, path, Tk_MainWindow (interp));
	
	if (tkwin == NULL) {
		/*The result should already be set*/
		return TCL_ERROR;
	}

	Tk_MakeWindowExist (tkwin);

	if (!XShapeQueryExtension (Tk_Display (tkwin), &eventBase, &errorBase)) {
		Tcl_SetResult (interp, "Your display doesn't appear to support the X shape extension.", TCL_STATIC);
		return TCL_ERROR;
	}
	
	wPtr = (ThemeCursor *) ckalloc (sizeof (ThemeCursor));
	
	wPtr->dis = Tk_Display (tkwin);
	wPtr->tkwin = tkwin;
	wPtr->xWin = Tk_WindowId (tkwin);
	wPtr->mainWin = XmuClientWindow (wPtr->dis, wPtr->xWin);
	wPtr->cursor = cursor;
	wPtr->pix = None;
	wPtr->cursorGC = XCreateGC (wPtr->dis, wPtr->xWin, 0, 0);
	wPtr->screen = Tk_ScreenNumber (tkwin);
	wPtr->rootWin = DefaultRootWindow (wPtr->dis);
	wPtr->width = 1;
	wPtr->height = 1;
	wPtr->mapped = 0;
	wPtr->depth = Tk_Depth (tkwin);

	if (XpmCreatePixmapFromBuffer (wPtr->dis, wPtr->xWin, xpmData, &wPtr->pix, &pixShape, NULL)) {
		Tcl_SetResult (interp, "bad xpm", TCL_STATIC);
		ckfree ((char *)wPtr);
		return TCL_ERROR;
	}
	
	if (wPtr->pix == NULL) {
		Tcl_SetResult (interp, "invalid xpm pixmap", TCL_STATIC);
		ckfree ((char *)wPtr);
		return TCL_ERROR;
	}
	
	ThemeCursorCmdWindowCreateCursor (wPtr);

	cursor = Tk_GetCursorFromData (interp, tkwin, dummyCursorData, "", 1, 1, 1, 1, "black", "black");
	
	if (cursor == None) {
		/*The result should already be set*/
		ckfree ((char *)wPtr);
		return TCL_ERROR;
	}
	
	XShapeCombineMask (wPtr->dis, wPtr->cursorXWin, ShapeBounding, 0, 0, pixShape, ShapeSet);
	
	Tk_UndefineCursor (tkwin);
	Tk_DefineCursor (tkwin, cursor);
	Tk_CreateEventHandler (tkwin, PointerMotionMask, ThemeCursorEventMotion, (ClientData) wPtr);
	Tk_CreateEventHandler (tkwin, LeaveWindowMask, ThemeCursorEventLeave, (ClientData) wPtr);
	ThemeCursorGetMotionFromAllChildren (wPtr, wPtr->xWin);
	
	return TCL_OK;
}
